
##-------------------------------------------------------------------------------
## Copyright (c) 2012 University of Illinois, NCSA.
## All rights reserved. This program and the accompanying materials
## are made available under the terms of the 
## University of Illinois/NCSA Open Source License
## which accompanies this distribution, and is available at
## http://opensource.ncsa.illinois.edu/license.html
##-------------------------------------------------------------------------------
##--------------------------------------------------------------------------------------------------#
##' Convert BioCro output to netCDF
##'
##' Converts BioCro output to netCDF.
##' Modified from on model2netcdf.SIPNET and model2netcdf.ED2 by
##' @name model2netcdf.BIOCRO
##' @title Function to convert biocro model output to standard netCDF format
##' @param outdir Location of model output
##' @param lat Latitude of the site
##' @param lon Longitude of the site
##' @export
##' @author David LeBauer, Deepak Jaiswal, Rob Kooper
model2netcdf.BIOCRO <- function(result, genus = NULL, outdir, lat = -9999, lon = -9999) {
  

  require(data.table)
  require(lubridate)
  require(ncdf4)
  
  if(!is.data.table(result)){
    if(is.character(result)){
      if(file.exists(result)){load(result)}}}
  
  if(!("hour" %in% colnames(result))){
    result$hour <- 0
  }
  if(all(c("year", "hour", "doy") %in% colnames(result))){
    setnames(result, c("year", "hour", "doy"), c("Year", "Hour", "DayofYear"))
  }
  
  ## longname prefix station_* used for a point
  ## http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/cf-conventions.html#scalar-coordinate-variables
  x <- ncdim_def("latitude", "degrees_east",
                 vals =  as.numeric(lat),
                 longname = "station_latitude",
                 unlim = TRUE) 
  y <- ncdim_def("longitude", "degrees_north",
                 vals = as.numeric(lon),
                 longname = "station_longitude",
                 unlim = TRUE)

  for(yeari in unique(result$Year)){
    dates <- ymd(paste0(result$Year, "-01-01")) + days(as.numeric(result$DayofYear - 1)) + hours(result$Hour)
    days_since_origin <- ymd_hms(dates) - ymd_hms("1700-01-01 00:00:00")
    if(!units(days_since_origin) == 'days') stop('check time units')
    t <- ncdim_def("time", "days since 1700-01-01", 
                   as.numeric(days_since_origin)) #define netCDF dimensions for variables
    if(exists('genus') & (genus == "Saccharum")){
      for(variable in c("Leaf", "Root", "Stem", "LAI", "DayofYear")) {
        v <- result[[variable]]
        result[[variable]] <- c(v[1], rep(v[-1], 24, each = TRUE))
      }
    }
    
    vars <- list()
    
    c2biomass <- 0.4
    vars <- list(
      TotLivBiom  = mstmipvar("TotLivBiom", x, y, t),
      RootBiom    = mstmipvar("RootBiom", x, y, t),
      StemBiom    = mstmipvar("StemBiom", x, y, t),
      Yield       = mstmipvar("Yield", x, y, t),
      Evap        = mstmipvar("Evap", x, y, t),
      TVeg        = mstmipvar("TVeg", x, y, t),
      LAI         = mstmipvar("LAI", x, y, t))
    
    k <- ud.convert(1, "Mg/ha", "kg/m2") / c2biomass
    
    results <- with(result, list(
      TotLivBiom = k * (Leaf + Root + Stem + Rhizome + Grain),
      RootBiom = k * Root,
      StemBiom = k * Stem,
      Yield = Stem,
      Evap = ud.convert(SoilEvaporation + CanopyTrans, "Mg/ha/h", "kg/m2/s"),
      TVeg = ud.convert(CanopyTrans, "Mg/ha/h", "kg/m2/s"),
      LAI =  LAI))
    
    ncfile <- file.path(outdir, paste0(yeari, ".nc"))
    if(file.exists(ncfile)){
      nc <- nc_open(ncfile, write = TRUE)
    } else {
      nc <- nc_create(filename = file.path(outdir, paste0(yeari, ".nc")),
                      vars = vars)      
      ncatt_put(nc, 0, "description",
                "This is an output from the BioCro Crop model generated by the model2netcdf.BIOCRO.R function in the PEcAn.BIOCRO package; see github.com:PecanProject/pecan/wiki for more information")
    }    
    
    varfile <- file(file.path(outdir, paste(yeari, "nc", "var", sep=".")), "w")
    
    ## Output netCDF data
    for(.vname in names(vars)) {
      ncvar_put(nc, varid = vars[[.vname]], vals = results[[.vname]])
      cat(paste(vars[[.vname]]$name, vars[[.vname]]$longname), file=varfile, sep="\n")
    }
    
    close(varfile)
    nc_close(nc)
  }
  
}

####################################################################################################
### EOF.  End of R script file.              
####################################################################################################
